/******************************************************************************
 * Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden.
 *
 *                        All Rights Reserved
 *
 *               www.mindbright.se, info@mindbright.se
 *
 * This program and its source-code may be used FREE of charge for strictly
 * NON-COMMERCIAL purposes. Companies or persons wishing to make non-commercial
 * software that in whole or in part contains or is derived from this program
 * or any part thereof must inform the original author of such work. The
 * original author must also be given attribution as the author of used parts.
 * For COMMERCIAL use all parts of this program originally developed by
 * Mindbright Technology must be LICENSED under Mindbright Technology's
 * SHAREWARE license, please contact Mindbright Technology for information on
 * this. This copyright notice must allways be included and unchanged.
 *
 * IF LICENSED FREE OF CHARGE AND/OR IF NOT STATED OTHERWISE THIS PROGRAM COMES
 * WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, IN NO EVENT SHALL
 * MINDBRIGHT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.util.Vector;

import mindbright.security.*;

public class SSHChannelController extends SSH implements SSHChannelListener {

  protected SSHTxChannel        txChan;
  protected SSHRxChannel        rxChan;
  protected SSHConnectChannel   cnChan;
  protected SSHPduQueue         txQueue;
  protected SSHPduQueue         cnQueue;

  protected int    nextEmptyChan;
  protected Vector tunnels;

  protected Vector listenChannels;

  protected SSHClient client;
  protected SSHStdIO  console;

  protected Cipher sndCipher;
  protected Cipher rcvCipher;

  public SSHChannelController(SSHClient client, InputStream in, OutputStream out,
			      Cipher sndCipher, Cipher rcvCipher,
			      SSHStdIO console) {
    this.sndCipher = sndCipher;
    this.rcvCipher = rcvCipher;

    this.client  = client;
    this.console = console;

    this.tunnels         = new Vector();
    this.nextEmptyChan   = 0;
    this.listenChannels  = new Vector();
    
    txChan = new SSHTxChannel(out, MAIN_CHAN_NUM);
    rxChan = new SSHRxChannel(in, MAIN_CHAN_NUM);
    cnChan = new SSHConnectChannel(this);

    rxChan.setSSHChannelListener(this);
    txChan.setSSHChannelListener(this);
    cnChan.setSSHChannelListener(this);
    rxChan.setSSHPduFactory(new SSHPduInputStream(MSG_ANY, rcvCipher));
    txQueue = txChan.getQueue();
    cnQueue = cnChan.getQueue();
  }

  public void start() {
    txChan.start();
    rxChan.start();
    cnChan.start();
  }

  public void waitForExit() throws InterruptedException {
    rxChan.join();
    killListenChannels();
    if(txChan.isAlive())
      txChan.stop();
    if(cnChan.isAlive())
      cnChan.stop();
  }

  public synchronized int newChannelId() {
    int newChan = nextEmptyChan;
    if(nextEmptyChan < tunnels.size()) {
      int i;
      for(i = nextEmptyChan + 1; i < tunnels.size(); i++)
	if(tunnels.elementAt(i) == null)
	  break;
      nextEmptyChan = i;
    } else {
      tunnels.addElement(null);
      nextEmptyChan++;
    }
    return newChan;
  }

  public synchronized void addTunnel(SSHTunnel tunnel) throws IOException {
    tunnels.setElementAt(tunnel, tunnel.channelId);
  }

  public synchronized SSHTunnel delTunnel(int channelId) {
    SSHTunnel tunnelToDelete = (SSHTunnel) tunnels.elementAt(channelId);
    tunnels.setElementAt(null, channelId);
    nextEmptyChan = (channelId < nextEmptyChan ? channelId : nextEmptyChan);
    return tunnelToDelete;
  }

  public boolean haveHostInFwdOpen() {
    return client.isProtocolFlagSet(PROTOFLAG_HOST_IN_FWD_OPEN);
  }

  public SSHListenChannel newListenChannel(int localPort, String remoteHost, int remotePort,
					   String plugin) {
    SSHListenChannel newListenChan = null;
    try {
      newListenChan = SSHProtocolPlugin.getPlugin(plugin).localListener(localPort,
									remoteHost, remotePort,
									this);
      newListenChan.setSSHChannelListener(this);
      newListenChan.start();
      synchronized(listenChannels) {
	listenChannels.addElement(newListenChan);
      }
    } catch(IOException e) {
      // !!!
      System.out.println("newListenChannel: " + e.toString());
    }
    return newListenChan;
  }

  public void killListenChannel(int listenPort) {
    SSHListenChannel listenChan;
    synchronized(listenChannels) {
      for(int i = 0; i < listenChannels.size(); i++) {
	listenChan = (SSHListenChannel) listenChannels.elementAt(i);
	if(listenChan.getListenPort() == listenPort) {
	  listenChannels.removeElementAt(i);
	  if(listenChan.isAlive())
	    listenChan.stop();
	  break;
	}
      }
    }
  }

  public void killListenChannels() {
    SSHListenChannel listenChan;
    synchronized(listenChannels) {
      while(listenChannels.size() > 0) {
	listenChan = (SSHListenChannel) listenChannels.elementAt(0);
	if(listenChan.isAlive())
	  listenChan.stop();
	listenChannels.removeElementAt(0);
      }
    }
  }

  public SSHPdu prepare(SSHPdu pdu) {
    return pdu;
  }

  public void transmit(SSHPdu pdu) {
    txQueue.addLast(pdu);
  }

  public void receive(SSHPdu pdu) {
    SSHPduInputStream inPdu = (SSHPduInputStream) pdu;
    SSHTunnel         tunnel;
    int               channelNum;
    try {
      switch(inPdu.type) {
      case MSG_CHANNEL_DATA:
	channelNum  = inPdu.readInt();
	tunnel      = (SSHTunnel)tunnels.elementAt(channelNum);
	if(tunnel != null)
	  tunnel.transmit(pdu);
	else
	  throw new Exception("Data on nonexistent channel: " + channelNum);
	break;
      case SMSG_STDOUT_DATA:
      case SMSG_STDERR_DATA:
	if(console != null)
	  console.print(inPdu.readString());
	break;
      case SMSG_EXITSTATUS:
	SSHPduOutputStream exitPdu = new SSHPduOutputStream(CMSG_EXIT_CONFIRMATION, sndCipher);
	int status = inPdu.readInt();
	if(console != null) {
	  if(status != 0)
	    console.serverDisconnect(client.getServerName() + " disconnected: " + status);
	  else
	    console.serverDisconnect("Connection to " + client.getServerName() + " closed.");
	}
	transmit(exitPdu);
	break;
      case MSG_PORT_OPEN:
	cnQueue.addLast(inPdu);
	break;
      case MSG_CHANNEL_OPEN_CONFIRMATION:
	channelNum  = inPdu.readInt();
	tunnel = (SSHTunnel)tunnels.elementAt(channelNum);
	if(tunnel != null) {
	  if(!tunnel.setRemoteChannelId(inPdu.readInt()))
	    throw new Exception("Open confirmation on allready opened channel!");
	  tunnel.start();
	} else
	  throw new Exception("Open confirm on nonexistent: " + channelNum);
	break;
      case MSG_CHANNEL_OPEN_FAILURE:
	channelNum = inPdu.readInt();
	if(delTunnel(channelNum) != null)
	  throw new Exception("Open failure on nonexistent channel: " + channelNum);
	break;
      case MSG_CHANNEL_INPUT_EOF:
	channelNum  = inPdu.readInt();
	tunnel      = (SSHTunnel)tunnels.elementAt(channelNum);
	if(tunnel != null) {
	  tunnel.receiveInputEOF();
	} else
	  throw new Exception("Input eof on nonexistent channel: " + channelNum);
	break;
      case MSG_CHANNEL_OUTPUT_CLOSED:
	channelNum  = inPdu.readInt();
	tunnel      = (SSHTunnel)tunnels.elementAt(channelNum);
	if(tunnel != null) {
	  tunnel.receiveOutputClosed();
	} else
	  throw new Exception("Output closed on nonexistent channel: " + channelNum);
	break;
      case MSG_DISCONNECT:
	throw new Exception("Server disconnected: " + inPdu.readString());
      default:
	throw new Exception("Unknown packet type (" + inPdu.type + "), disconnecting...");
      }
    } catch (Exception e) {
      // !!!
      sendDisconnect("error in mainchannel: " + e.getMessage());
    }
  }

  public void close(SSHChannel chan) {
    // !!!
    if(chan instanceof SSHConnectChannel)
      System.out.println("Controller connect-channel closed");
    else if(chan instanceof SSHTxChannel)
      System.out.println("Controller TX-channel closed");
    else if(chan instanceof SSHRxChannel)
      System.out.println("Controller RX-channel closed");
    else if(chan instanceof SSHListenChannel)
      System.out.println("Listen channel for port " + ((SSHListenChannel)chan).getListenPort() +
			 " closed");
    else
      System.out.println("ERROR: must be a bug !!!");
  }

  public void sendDisconnect(String reason) {
    try {
      SSHPduOutputStream pdu = new SSHPduOutputStream(MSG_DISCONNECT, sndCipher);
      pdu.writeString(reason);
      txQueue.addFirst(pdu);
      Thread.sleep(500);
      txChan.setClosePending();
      if(console != null)
	console.serverDisconnect("\r\nDisconnecting, " + reason);
    } catch (Exception e) {
      // !!!
      System.out.println("Error in sendDisconnect: " + e.toString());
    }
  }

  public void addHostMapTemporary(String fromHost, String toHost, int toPort) {
    cnChan.addHostMapTemporary(fromHost, toHost, toPort);
  }

  public void addHostMapPermanent(String fromHost, String toHost, int toPort) {
    cnChan.addHostMapPermanent(fromHost, toHost, toPort);
  }

  public void delHostMap(String fromHost) {
    cnChan.delHostMap(fromHost);
  }

  public Vector getHostMap(String fromHost) {
    return cnChan.getHostMap(fromHost);
  }

}
