/******************************************************************************
 * Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden.
 *
 *                        All Rights Reserved
 *
 *               www.mindbright.se, info@mindbright.se
 *
 * This program and its source-code may be used FREE of charge for strictly
 * NON-COMMERCIAL purposes. Companies or persons wishing to make non-commercial
 * software that in whole or in part contains or is derived from this program
 * or any part thereof must inform the original author of such work. The
 * original author must also be given attribution as the author of used parts.
 * For COMMERCIAL use all parts of this program originally developed by
 * Mindbright Technology must be LICENSED under Mindbright Technology's
 * SHAREWARE license, please contact Mindbright Technology for information on
 * this. This copyright notice must allways be included and unchanged.
 *
 * IF LICENSED FREE OF CHARGE AND/OR IF NOT STATED OTHERWISE THIS PROGRAM COMES
 * WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, IN NO EVENT SHALL
 * MINDBRIGHT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.math.BigInteger;
import java.security.SecureRandom;

import mindbright.security.*;
import mindbright.util.*;

public class SSHPduOutputStream extends DataOutputStream implements SSHPdu {

  static class PduByteArrayOutputStream extends ByteArrayOutputStream {
    PduByteArrayOutputStream() {
      super();
    }

    PduByteArrayOutputStream(int size) {
      super(size);
    }

    public byte[] getBuf() {
      return buf;
    }

    public int getCount() {
      return count;
    }

    public void setBuf(byte[] buf) {
      this.buf = buf;
    }

    public void setCount(int count) {
      this.count = count;
    }
  }

  public static final int SSH_DEFAULT_PKT_LEN = 8192;

  byte[]  readFromRawData;
  int     readFromOff;
  int     readFromSize;

  public int    type;
  public Cipher cipher;

  SSHPduOutputStream(Cipher cipher) {
    super(null);
    this.cipher = cipher;
  }

  SSHPduOutputStream(int type, Cipher cipher) throws IOException {
    super(new PduByteArrayOutputStream(SSH_DEFAULT_PKT_LEN));
    this.type   = type;
    this.cipher = cipher;
    if(cipher != null) {
      SecureRandom rand = new SecureRandom();
      byte[] randBytes  = new byte[8];
      rand.nextBytes(randBytes);
      for(int i = 0; i < 8; i++)
	write(randBytes[i]);
    } else {
      for(int i = 0; i < 8; i++)
	write(0);
    }
    write(type);
  }

  public SSHPdu createPdu() {
    SSHPdu pdu;
    try {
      pdu = new SSHPduOutputStream(this.type, this.cipher);
    } catch (IOException e) {
      // !!!
      pdu = null;
    }
    return pdu;
  }

  public void writeBigInteger(BigInteger bi) throws IOException {
    short bytes = (short)((bi.bitLength() + 7) / 8);
    byte[] raw  = bi.toByteArray();
    writeShort(bi.bitLength());
    if(raw[0] == 0)
      write(raw, 1, bytes);
    else
      write(raw, 0, bytes);
  }

  public void writeString(String str) throws IOException {
    byte[] raw = str.getBytes();
    writeInt(raw.length);
    write(raw);
  }

  public void readFrom(InputStream in) throws IOException {
    if(type != SSH.MSG_CHANNEL_DATA &&
       type != SSH.CMSG_STDIN_DATA)
      throw new IOException("Trying to read raw data into non-data PDU");

    PduByteArrayOutputStream bytes = (PduByteArrayOutputStream) out;

    readFromRawData = bytes.getBuf();
    readFromOff     = bytes.size() + 4; // Leave space for size

    readFromSize = in.read(readFromRawData, readFromOff, SSH_DEFAULT_PKT_LEN - readFromOff);
    if(readFromSize == -1)
      throw new IOException("EOF");

    writeInt(readFromSize);
    bytes.setCount(readFromOff + readFromSize);
  }

  public void writeTo(OutputStream sshOut) throws IOException {
    PduByteArrayOutputStream bytes   = (PduByteArrayOutputStream) out;
    int                      pad     = (bytes.size() + 4) % 8;
    byte[]                   padData = new byte[bytes.size() + 4 - pad];
    int                      crc32   = (int)CRC32.getValue(bytes.getBuf(), pad, bytes.size() - pad);

    writeInt(crc32);
    System.arraycopy(bytes.getBuf(), pad, padData, 0, padData.length);

    if(cipher != null) {
      padData = cipher.encrypt(padData);
    }

    bytes = new PduByteArrayOutputStream();
    bytes.setBuf(padData);
    bytes.setCount(padData.length);
    out = bytes;
    sshOut = new DataOutputStream(sshOut);

    ((DataOutputStream)sshOut).writeInt(bytes.getCount() - (8 - pad));
    bytes.writeTo(sshOut);
    sshOut.flush();
  }

  public byte[] rawData() {
    return readFromRawData;
  }
  public void rawSetData(byte[] raw) {
  }
  public int rawOffset() {
    return readFromOff;
  }
  public int rawSize() {
    // !!! return readFromSize;
    //
    byte[] bytes = readFromRawData;
    int    off   = readFromOff - 4;
    int ch1 = bytes[off++];
    int ch2 = bytes[off++];
    int ch3 = bytes[off++];
    int ch4 = bytes[off];
    return ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0));
  }
  public void rawAdjustSize(int size) {
    PduByteArrayOutputStream bytes = (PduByteArrayOutputStream) out;
    bytes.setCount(size);
  }

}

