/******************************************************************************
 * Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden.
 *
 *                        All Rights Reserved
 *
 *               www.mindbright.se, info@mindbright.se
 *
 * This program and its source-code may be used FREE of charge for strictly
 * NON-COMMERCIAL purposes. Companies or persons wishing to make non-commercial
 * software that in whole or in part contains or is derived from this program
 * or any part thereof must inform the original authors of such work. The
 * original author must also be given attribution as the author of used parts.
 * For COMMERCIAL use all parts of this program originally developed by
 * Mindbright Technology must be LICENSED under Mindbright Technology's
 * SHAREWARE license, please contact Mindbright Technology for information on
 * this. This copyright notice must allways be included and unchanged.
 *
 * IF LICENSED FREE OF CHARGE AND/OR IF NOT STATED OTHERWISE THIS PROGRAM COMES
 * WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, IN NO EVENT SHALL
 * MINDBRIGHT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *****************************************************************************/

/*
 * TODO: o Reverse autowrap
 *       o The keyKludge* is not too nice, general behaviour wanted (e.g. what to send on BS etc.)
 *       o Colors are not supported yet
 *       o Cursor keys are not implemented (also <ESC>[A... or <ESC>OA toggle)
 *       o Event-multicaster (for TerminalListener)
 *       o Application keypad not supported
 *       o ALT_GR on WinLost does some funny things...
 *       o General cleanup of code
 *       o Many things are not implemented for VT100/xterm (see TerminalVT100.java)
 *       o etc. etc. etc. (I tend to foreget :-)
 */

package mindbright.terminal;

import java.awt.*; 
import java.awt.event.*; 

public class TerminalWin extends Canvas
  implements Terminal, KeyListener, AdjustmentListener {

  boolean enterKeyKludge = false;
  boolean metaKeyKludge  = false;
  boolean ctrlKeyKludge  = false;
  boolean bsKeyKludge    = false;

  final static boolean DEBUG = false;

  final static public int GRAVITY_SOUTHWEST = 0;
  final static public int GRAVITY_NORTHWEST = 1;

  final static public int MIN_ROWS = 2;
  final static public int MIN_COLS = 8;
  final static public int MAX_COLS = 256;
  final static public int MAX_ROWS = 256;

  final static public int MAJOR_DIRTY_AREA  = 240;
  final static public int UPDATE_DIRTY_AREA = 2;

  TerminalListener    listener;
  TerminalInterpreter interpreter;
  Scrollbar           scrollbar;

  TerminalWin self;

  boolean repaintPending;

  boolean complexScroll;

  int     dirtyCount;
  int     dirtyTop;
  int     dirtyBottom;
  int     dirtyLeft;
  int     dirtyRight;
  boolean fullRefresh;

  boolean lineWrap;
  boolean reverseVideo;
  boolean autoCR;
  boolean localEcho;
  int     resizeGravity;
  boolean autoRepositionOutput;
  boolean autoRepositionInput;

  boolean cursorKeysMode;

  int rows;
  int cols;
  int vpixels;
  int hpixels;
  int borderWidth  = 2;
  int borderHeight = 2;

  int windowTop;
  int windowBottom;
  int windowLeft;
  int windowRight;

  int charWidth;
  int charHeight;
  int charMaxAscent;
  int charMaxDescent;
  int charLeading;

  int curRow;
  int curCol;
  int lastWriteRow;
  int lastWriteCol;
  int lastCursorRow;
  int lastCursorCol;

  int     selectRowAnchor;
  int     selectColAnchor;
  int     selectRowLast;
  int     selectColLast;
  boolean hasSelection;
  boolean selectReverse;

  int curAttr;

  int curRowSave;
  int curColSave;
  int curAttrSave;

  Color bgColor;
  Color fgColor;
  Color origBgColor;
  Color origFgColor;

  public final static Color termColors[] = {
    Color.red,
    Color.pink,
    Color.orange,
    Color.yellow,
    Color.green,
    Color.magenta,
    Color.cyan,
    Color.blue
  };

  char[][]   screen;
  int[][]    attributes;

  int saveLines;
  int visTop;

  public final static int ATTR_CHARNOTDRAWN = 0x0000;
  public final static int ATTR_SELECTED     = 0x1000;
  public final static int ATTR_CHARDRAWN    = 0x8000;
  public final static int ATTR_BOLD         = 0x0001;
  public final static int ATTR_LOWINTENSITY = 0x0002;
  public final static int ATTR_UNDERLINE    = 0x0004;
  public final static int ATTR_BLINKING     = 0x0008;
  public final static int ATTR_INVERSE      = 0x0010;
  public final static int ATTR_INVISIBLE    = 0x0020;
  public final static int ATTR_FGCOLOR      = 0x0040;
  public final static int ATTR_BGCOLOR      = 0x0080;

  public final static int IDX_ATTR  = 0;
  public final static int IDX_FGCOL = 1;
  public final static int IDX_BGCOL = 2;
  public final static int IDX_LAST  = 3;

  public final static char[] spacerow;
  public final static int[]  zerorow;

  static {
    spacerow = new char[MAX_COLS];
    zerorow  = new int[MAX_COLS * IDX_LAST];
    for(int i = 0; i < MAX_COLS; i++) {
      spacerow[i] = ' ';
      for(int j = 0; j < IDX_LAST; j++)
	zerorow[(i * IDX_LAST) + j] = 0;
    }
  }

  Image     memImage;
  Graphics  memGraphics;
  Dimension memImageSize;

  public Font plainFont;
  public Font boldFont;

  public TerminalWin(int width, int height, TerminalInterpreter interpreter) {
    super();

    self = this;

    curAttr              = ATTR_CHARDRAWN;
    curRow               = 0;
    curCol               = 0;
    lineWrap             = true;
    reverseVideo         = false;
    autoCR               = false;
    localEcho            = false;
    resizeGravity        = GRAVITY_SOUTHWEST;
    autoRepositionOutput = true;
    autoRepositionInput  = true;

    saveLines     = 0;
    visTop        = 0;

    // !!! 
    repaintPending = true;

    setFont("Monospaced", 12);

    dirtyTop    = 0;
    dirtyBottom = rows;
    dirtyLeft   = 0;
    dirtyRight  = cols;

    setWindowSize(height, width);
    resetWindow();
    clearScreen();

    if(interpreter == null)
      interpreter = new TerminalDumb();
    this.interpreter = interpreter;
    interpreter.setTerminal(this);

    origBgColor = Color.white;
    origFgColor = Color.black;

    setForeground(origFgColor);
    setBackground(origBgColor);

    scrollbar = null;

    addKeyListener(this);
    addComponentListener(new ComponentAdapter() {
      public void componentResized(ComponentEvent e)  { self.componentResized(e);  }
    });
    addMouseMotionListener(new MouseMotionAdapter() {
      public void mouseDragged(MouseEvent e)  { self.mouseDragged(e);  }
    });
    addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent e)  { self.mousePressed(e);  }
    });
  }

  public void setFont(String name, int size) {

    plainFont = new Font(name, Font.PLAIN, size);
    boldFont  = new Font(name, Font.BOLD, size);

    super.setFont(plainFont);

    if(isShowing()) {
      getDimensionOfText(0, 0);
      componentResized(null);
    }

  }

  public void setFont(Font font) {
    setFont(font.getName(), font.getSize());
  }

  void updateScrollbarValues() {
    if(scrollbar != null) {
      scrollbar.setMinimum(0);
      scrollbar.setMaximum(saveLines + rows);
      scrollbar.setVisibleAmount(rows);
      scrollbar.setBlockIncrement(rows);
      scrollbar.setValue(visTop);
    }
  }

  public Scrollbar createScrollbar() {
    if(scrollbar != null)
      return scrollbar;
    scrollbar = new Scrollbar(Scrollbar.VERTICAL);
    updateScrollbarValues();
    scrollbar.addAdjustmentListener(this);
    return scrollbar;
  }

  public Scrollbar getScrollbar() {
    return scrollbar;
  }

  public void setScrollbar(Scrollbar scrollbar) {
    if(this.scrollbar != null)
      this.scrollbar.removeAdjustmentListener(this);
    this.scrollbar = scrollbar;
    scrollbar.addAdjustmentListener(this);
    updateScrollbarValues();
  }

  public void setSaveLines(int n) {
    int oldSaveLines = saveLines;
    int fromRow, toRow, copyRows;
    n = (n < 0 ? 0 : n);
    n = (n > 32168 ? 32168 : n);
    if(saveLines != n) {
      char[][] oldScreen     = screen;
      int[][]  oldAttributes = attributes;
      saveLines              = n;
      visTop                 = n;
      setWindowSize(rows, cols);
      if(oldSaveLines < saveLines) {
	fromRow  = 0;
	toRow    = (saveLines - oldSaveLines);
	copyRows = oldSaveLines + rows;
      } else {
	fromRow  = (oldSaveLines - saveLines);
	toRow    = 0;
	copyRows = saveLines + rows;
      }
      System.arraycopy(oldScreen, fromRow, screen, toRow, copyRows);
      System.arraycopy(oldAttributes, fromRow, attributes, toRow, copyRows);
      updateScrollbarValues();
    }
  }

  public int getSaveLines() {
    return saveLines;
  }

  public void setAutoRepositionOnInput(boolean val) {
    autoRepositionInput = val;
  }

  public void setAutoRepositionOnOutput(boolean val) {
    autoRepositionOutput = val;
  }

  public boolean getAutoRepositionOnInput() {
    return autoRepositionInput;
  }

  public boolean getAutoRepositionOnOutput() {
    return autoRepositionOutput;
  }

  public void setCursorKeysMode(boolean val) {
    cursorKeysMode = val;
  }

  public void setWindowSize(int rows, int cols) {
    this.rows    = rows;
    this.cols    = cols;
    screen       = new char[rows + saveLines][cols];
    attributes   = new int[rows + saveLines][cols * IDX_LAST];
  }

  public void setInterpreter(TerminalInterpreter interpreter) {
    if(interpreter != null)
      this.interpreter = interpreter;
  }

  public void setResizeGravity(int resizeGravity) {
    this.resizeGravity = resizeGravity;
  }

  public void setReverseVideo(boolean val) {
    Color swap;
    if(val != reverseVideo) {
      reverseVideo = val;
      swap         = origBgColor;
      origBgColor  = origFgColor;
      origFgColor  = swap;
      makeAllDirty();
    }
  }

  public boolean getReverseVideo() {
    return reverseVideo;
  }

  public void setLocalEcho(boolean val) {
    localEcho = val;
  }

  public void setAutoCR(boolean val) {
    autoCR = val;
  }

  public void doBell() {
    Toolkit toolkit = getToolkit();
    if(toolkit != null)
      toolkit.beep();
  }

  public void doBS() {
    curCol--;
    if(curCol < windowLeft) {
      curCol = windowLeft;
    }
    if(DEBUG) System.out.println("doBS");
  }

  public void doTab() {
    int end = curCol + (8 - (curCol % 8));
    if(end > windowRight)
      end = windowRight;
    curCol = end;
    if(DEBUG) System.out.println("doTab");
  }

  public void doCR() {
    curCol = windowLeft;
    if(DEBUG) System.out.println("doCR");
  }

  public void doLF() {
    curRow += 1;
    if(autoCR)
      doCR();
    if(DEBUG) System.out.println("doLF");
  }

  public boolean getAutoWrap() {
    return lineWrap;
  }
  public void setAutoWrap(boolean val) {
    lineWrap = val;
  }

  public void resetWindow() {
    windowTop     = 0;
    windowBottom  = rows;
    windowLeft    = 0;
    windowRight   = cols;
    complexScroll = false;
  }
  public void setWindow(int top, int bottom) {
    setWindow(top, 0, bottom, cols);
  }
  public void setWindow(int top, int left, int bottom, int right) {
    windowTop    = top;
    windowLeft   = left;
    windowBottom = bottom;
    windowRight  = right;
    if(DEBUG) System.out.println("setWindow: " + top + ", " + bottom + ", " + left + ", " + right);


    // Ensure that the selected area is totally outside the scrolling
    // region OR that the scrolling region starts at the top of the
    // screen and the selection is completely above the scrolling
    // regions bottom. This makes things alot easier and is not that much
    // of problem.
    //
    if(hasSelection) {
      int selRowAnch = selectRowAnchor - visTop;
      int selRowLast = selectRowLast - visTop;
      if(top != 0 && (selRowAnch >= 0 || selRowLast >= 0)) {
	if(!(selRowAnch < top && selRowLast < top ||
	     selRowAnch >= bottom && selRowLast >= bottom))
	  clearSelection();
      } else {
	if(!(selRowAnch < bottom && selRowLast < bottom))
	  clearSelection();
      }
    }

    if(windowLeft != 0 || windowRight != cols)
      complexScroll = true;
    else
      complexScroll = false;

  }
  public int  getWindowTop() {
    return windowTop;
  }
  public int  getWindowBottom() {
    return windowBottom;
  }
  public int  getWindowLeft() {
    return windowLeft;
  }
  public int  getWindowRight() {
    return windowRight;
  }


  public int getCursorV() {
    return curRow;
  }
  public int getCursorH() {
    return curCol;
  }

  public void cursorSetPos(int v, int h, boolean relative) {
    if(DEBUG) System.out.println("cursorSetPos: " + v + ", " + h + "(" + relative + ")");
    int maxV = rows - 1;
    int maxH = cols - 1;
    int minV = 0;
    int minH = 0;
    if(relative) {
      v += windowTop;
      maxV = windowBottom - 1;
      minV = windowTop;
      h += windowLeft;
      maxH = windowRight - 1;
      minH = windowLeft;
    }
    if(v < minV)
      v = minV;
    if(h < minH)
      h = minH;
    if(v > maxV)
      v = maxV;
    if(h > maxH)
      h = maxH;
    curRow = v;
    curCol = h;
  }
  public void cursorUp(int n) {
    if(DEBUG) System.out.println("cursorUp: " + n);
    int min = (curRow < windowTop ? 0 : windowTop);
    curRow -= n;
    if(curRow < min)
      curRow = min;
  }
  public void cursorDown(int n) {
    if(DEBUG) System.out.println("cursorDown: " + n);
    int max = (curRow > windowBottom - 1 ? rows - 1: windowBottom - 1);
    curRow += n;
    if(curRow > max)
      curRow = max;
  }
  public void cursorForward(int n) {
    if(DEBUG) System.out.println("cursorFwd: " + n);
    curCol += n;
    if(curCol > windowRight)
      curCol = windowRight;
  }
  public void cursorBackward(int n) {
    if(DEBUG) System.out.println("cursorBack: " + n);
    curCol -= n;
    if(curCol < windowLeft)
      curCol = windowLeft;
  }
  public void cursorIndex(int n) {
    if(DEBUG) System.out.println("cursorIndex: " + n);
    if(curRow > windowBottom || curRow + n <= windowBottom)
      cursorDown(n);
    else {
      int m = windowBottom - curRow;
      cursorDown(m);
      scrollUp(n - m);
    }
  }
  public void cursorIndexRev(int n) {
    if(DEBUG) System.out.println("cursorIndexRev: " + n);
    if(curRow < windowTop || curRow - n >= windowTop)
      cursorUp(n);
    else {
      int m = curRow - windowTop;
      scrollDown(n - m);
      cursorUp(m);
    }
  }

  public void cursorSave() {
    curRowSave  = curRow;
    curColSave  = curCol;
    curAttrSave = curAttr;
  }
  public void cursorRestore() {
    curRow  = curRowSave;
    curCol  = curColSave;
    curAttr = curAttrSave;
  }

  public void scrollUp(int n) {
    int windowHeight = windowBottom - windowTop;
    int i, j = windowTop;

    if(DEBUG) System.out.println("scrollUp: " + n);

    if(complexScroll) {
      if(n < windowHeight) {
	j = (windowHeight - n) + windowTop;
	for(i = windowTop; i < j; i++) {
	  System.arraycopy(screen[saveLines + i + n], windowLeft, screen[saveLines + i],
			   windowLeft, windowRight - windowLeft);
	  System.arraycopy(attributes[saveLines + i + n], windowLeft * IDX_LAST,
			   attributes[saveLines + i], windowLeft * IDX_LAST,
			   (windowRight - windowLeft) * IDX_LAST);
	}
      }
      for(i = j; i < windowBottom; i++) {
	System.arraycopy(spacerow, 0, screen[saveLines + i], windowLeft, windowRight - windowLeft);
	System.arraycopy(zerorow, 0, attributes[saveLines + i], windowLeft * IDX_LAST,
			 (windowRight - windowLeft) * IDX_LAST);
      }
    } else {
      if(hasSelection) {
	if(windowTop == 0 && !(((selectRowAnchor - n) < 0) ||
			       ((selectRowLast   - n) < 0))) {
	  selectRowAnchor -= n;
	  selectRowLast -= n;
	} else {
	  clearSelection();
	}
      }
      if(windowTop == 0 && saveLines > 0) {
	int sl = (n < windowHeight ? n : windowHeight);
	int ll = windowHeight - sl;
	System.arraycopy(screen, sl, screen, 0, saveLines + ll);
	System.arraycopy(attributes, sl, attributes, 0, saveLines + ll);
	for(i = windowHeight - sl; i < windowHeight; i++) {
	  screen[saveLines + i]     = new char[cols];
	  attributes[saveLines + i] = new int[cols * IDX_LAST];
	}
      } else {
	if(n < windowHeight) {
	  j = (windowHeight - n) + windowTop;
	  System.arraycopy(screen, saveLines + windowTop + n, screen, saveLines + windowTop,
			   (windowHeight - n));
	  System.arraycopy(attributes, saveLines + windowTop + n, attributes,
			   saveLines + windowTop, (windowHeight - n));
	}
	for(i = j; i < windowBottom; i++) {
	  screen[saveLines + i]     = new char[cols];
	  attributes[saveLines + i] = new int[cols * IDX_LAST];
	}
      }
    }

    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(windowTop, windowLeft, windowBottom, windowRight);
  }
  public void scrollDown(int n) {
    int windowHeight = windowBottom - windowTop;
    int i, j = windowBottom;

    if(DEBUG) System.out.println("scrollDown: " + n);

    if(complexScroll) {
      if(n < windowHeight) {
	j = windowTop + n;
	for(i = windowBottom - 1; i >= j; i--) {
	  System.arraycopy(screen[saveLines + i - n], windowLeft, screen[saveLines + i], windowLeft, windowRight - windowLeft);
	  System.arraycopy(attributes[saveLines + i - n], windowLeft * IDX_LAST,
			   attributes[saveLines + i], windowLeft * IDX_LAST,
			   (windowRight - windowLeft) * IDX_LAST);
	}
      }
      for(i = windowTop; i < j; i++) {
	System.arraycopy(spacerow, 0, screen[saveLines + i], windowLeft, windowRight - windowLeft);
	System.arraycopy(zerorow, 0, attributes[saveLines + i], windowLeft * IDX_LAST,
			 (windowRight - windowLeft) * IDX_LAST);
      }
    } else {
      if(n < windowHeight) {
	j = windowTop + n;
	System.arraycopy(screen, saveLines + windowTop, screen, saveLines + windowTop + n,
			 windowHeight - n);
	System.arraycopy(attributes, saveLines + windowTop, attributes, saveLines + windowTop + n,
			 windowHeight - n);
      }
      for(i = windowTop; i < j; i++) {
	screen[saveLines + i]     = new char[cols];
	attributes[saveLines + i] = new int[cols * IDX_LAST];
      }
    }
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(windowTop, 0, windowBottom, cols);
  }

  public void clearBelow() {
    if(DEBUG) System.out.println("clearBelow");
    clearRight();
    for(int i = curRow + 1; i < windowBottom; i++) {
      screen[saveLines + i]     = new char[cols];
      attributes[saveLines + i] = new int[cols * IDX_LAST];
    }
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(curRow, 0, windowBottom, cols);
  }
  public void clearAbove() {
    if(DEBUG) System.out.println("clearAbove");
    clearLeft();
    for(int i = windowTop; i < curRow; i++) {
      screen[saveLines + i]     = new char[cols];
      attributes[saveLines + i] = new int[cols * IDX_LAST];
    }
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(windowTop, 0, curRow, cols);
  }
  public void clearScreen() {
    if(DEBUG) System.out.println("clearScreen");
    for(int i = windowTop; i < windowBottom; i++) {
      screen[saveLines + i]     = new char[cols];
      attributes[saveLines + i] = new int[cols * IDX_LAST];
    }
    dirtyCount = UPDATE_DIRTY_AREA;

    // Don't call updateDirtyArea(0, 0, rows, cols);
    // we want the values reset instead of updated
    //
    dirtyTop = 0;
    dirtyBottom = rows;
    dirtyLeft = 0;
    dirtyRight = cols;
  }
  public void clearRight() {
    if(DEBUG) System.out.println("clearRight");
    System.arraycopy(spacerow, 0, screen[saveLines + curRow], curCol, cols - curCol);
    System.arraycopy(zerorow, 0, attributes[saveLines + curRow], curCol * IDX_LAST,
		     (cols - curCol) * IDX_LAST);
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(curRow, curCol, curRow + 1, cols);
  }
  public void clearLeft() {
    if(DEBUG) System.out.println("clearLeft");
    System.arraycopy(spacerow, 0, screen[saveLines + curRow], 0, curCol);
    System.arraycopy(zerorow, 0, attributes[saveLines + curRow], 0, curCol * IDX_LAST);
    dirtyCount  = UPDATE_DIRTY_AREA;
    dirtyLeft   = 0;
    updateDirtyArea(curRow, 0, curRow + 1, curCol);
  }
  public void clearLine() {
    if(DEBUG) System.out.println("clearLine");
    screen[saveLines + curRow]     = new char[cols];
    attributes[saveLines + curRow] = new int[cols * IDX_LAST];
    dirtyCount  = UPDATE_DIRTY_AREA;
    dirtyLeft   = 0;
    dirtyRight  = cols;
    updateDirtyArea(curRow, 0, curRow + 1, cols);
  }

  public void insertChars(int n) {
    int edge = windowRight;
    if(DEBUG) System.out.println("inserChars: " + n);
    if(curCol < windowLeft || curCol > windowRight)
      return;
    if((curCol + n) < windowRight) {
      edge = curCol +  n;
      System.arraycopy(screen[saveLines + curRow], curCol, screen[saveLines + curRow], edge,
		       windowRight - edge);
      System.arraycopy(attributes[saveLines + curRow], curCol * IDX_LAST,
		       attributes[saveLines + curRow], edge * IDX_LAST,
		       (windowRight - edge) * IDX_LAST);
    }
    System.arraycopy(spacerow, 0, screen[saveLines + curRow], curCol, edge - curCol);
    System.arraycopy(zerorow, 0, attributes[saveLines + curRow], curCol * IDX_LAST,
		     (edge - curCol) * IDX_LAST);
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(curRow, curCol, curRow + 1, windowRight);
  }
  public void deleteChars(int n) {
    int edge = curCol;
    if(DEBUG) System.out.println("deleteChars: " + n);
    if(curCol < windowLeft || curCol > windowRight)
      return;
    if((curCol + n) < windowRight) {
      edge = windowRight - n;
      System.arraycopy(screen[saveLines + curRow], curCol + n, screen[saveLines + curRow], curCol,
		       edge - curCol);
      System.arraycopy(attributes[saveLines + curRow], (curCol + n) * IDX_LAST,
		       attributes[saveLines + curRow], curCol * IDX_LAST,
		       (edge - curCol) * IDX_LAST);
    }
    System.arraycopy(spacerow, 0, screen[saveLines + curRow], edge, windowRight - edge);
    System.arraycopy(zerorow, 0, attributes[saveLines + curRow], edge * IDX_LAST,
		     (windowRight - edge) * IDX_LAST);
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(curRow, curCol, curRow + 1, windowRight);
  }
  public void insertLines(int n) {
    int i, edge = windowBottom;
    if(DEBUG) System.out.println("insertLines: " + n);

    if(curRow < windowTop || curRow > windowBottom)
      return;

    if(complexScroll) {
      if(curRow + n < windowBottom) {
	edge = curRow  + n;
	for(i = windowBottom - 1; i >= edge; i--) {
	  System.arraycopy(screen[saveLines + i - n], windowLeft, screen[saveLines + i], windowLeft,
			   windowRight - windowLeft);
	  System.arraycopy(attributes[saveLines + i - n], windowLeft * IDX_LAST, attributes[saveLines + i],
			   windowLeft * IDX_LAST,
			   (windowRight - windowLeft) * IDX_LAST);
	}
      }
      for(i = curRow; i < edge; i++) {
	System.arraycopy(spacerow, 0, screen[saveLines + i], windowLeft, windowRight - windowLeft);
	System.arraycopy(zerorow, 0, attributes[saveLines + i], windowLeft * IDX_LAST,
			 (windowRight - windowLeft) * IDX_LAST);
      }
    } else {
      if(curRow + n < windowBottom) {
	edge = curRow  + n;
	System.arraycopy(screen, saveLines + curRow, screen, saveLines + edge, windowBottom - edge);
	System.arraycopy(attributes, saveLines + curRow, attributes, saveLines + edge,
			 windowBottom - edge);
      }
      for(i = curRow; i < edge; i++) {
	screen[saveLines + i]     = new char[cols];
	attributes[saveLines + i] = new int[cols * IDX_LAST];
      }
    }

    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(curRow, 0, windowBottom, cols);
  }
  public void deleteLines(int n) {
    int i, edge = curRow;
    if(DEBUG) System.out.println("deleteLines: " + n);

    if(curRow < windowTop || curRow > windowBottom)
      return;

    if(complexScroll) {
      if(curRow + n < windowBottom) {
	edge = windowBottom - n - 1;
	for(i = curRow; i <= edge; i++) {
	  System.arraycopy(screen[saveLines + i + n], windowLeft, screen[saveLines + i], windowLeft,
			   windowRight - windowLeft);
	  System.arraycopy(attributes[saveLines + i + n], windowLeft * IDX_LAST, attributes[saveLines + i],
			   windowLeft * IDX_LAST,
			   (windowRight - windowLeft) * IDX_LAST);
	}
      }
      for(i = edge; i < windowBottom; i++) {
	System.arraycopy(spacerow, 0, screen[saveLines + i], windowLeft, windowRight - windowLeft);
	System.arraycopy(zerorow, 0, attributes[saveLines + i], windowLeft * IDX_LAST,
			 (windowRight - windowLeft) * IDX_LAST);
      }
    } else {
      if(curRow + n < windowBottom) {
	edge = windowBottom - n - 1;
	System.arraycopy(screen, saveLines + curRow + n, screen, saveLines + curRow,
			 edge - curRow + 1);
	System.arraycopy(attributes, saveLines + curRow + n, attributes, saveLines + curRow,
			 edge - curRow + 1);
      }
      for(i = edge; i < windowBottom; i++) {
	screen[saveLines + i]     = new char[cols];
	attributes[saveLines + i] = new int[cols * IDX_LAST];
      }
    }

    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(curRow, 0, windowBottom, cols);
  }

  public void clearAllAttributes() {
    if(DEBUG) System.out.println("clearAllAttributes");
    curAttr = ATTR_CHARDRAWN;
  }
  public void setBold(boolean val) {
    if(DEBUG) System.out.println("setBold: " + val);
    if(val)
      curAttr |= ATTR_BOLD;
    else
      curAttr &= ~ATTR_BOLD;
  }
  public void setLowIntensity(boolean val) {
    if(DEBUG) System.out.println("setLowIntensity: " + val);
    if(val)
      curAttr |= ATTR_LOWINTENSITY;
    else
      curAttr &= ~ATTR_LOWINTENSITY;
  }
  public void setUnderline(boolean val) {
    if(DEBUG) System.out.println("setUnderline: " + val);
    if(val)
      curAttr |= ATTR_UNDERLINE;
    else
      curAttr &= ~ATTR_UNDERLINE;
  }
  public void setHilite(boolean val) {
    if(DEBUG) System.out.println("setBlinking: " + val);
    if(val)
      curAttr |= ATTR_BLINKING;
    else
      curAttr &= ~ATTR_BLINKING;
  }
  public void setInverseVideo(boolean val) {
    if(DEBUG) System.out.println("setInverseVideo: " + val);
    if(val)
      curAttr |= ATTR_INVERSE;
    else
      curAttr &= ~ATTR_INVERSE;
  }
  public void setInvisible(boolean val) {
    if(DEBUG) System.out.println("setInvisible: " + val);
    if(val)
      curAttr |= ATTR_INVISIBLE;
    else
      curAttr &= ~ATTR_INVISIBLE;
  }

  public void setForegroundColor(int c) {
    if(DEBUG) System.out.println("setForegroundColor: " + c);
    if(c >= 0 && c < 8) {
      fgColor = termColors[c];
      curAttr |= ATTR_FGCOLOR;
    }
  }
  public void setBackgroundColor(int c) {
    if(DEBUG) System.out.println("setBackgroundColor: " + c);
    if(c >= 0 && c < 8) {
      bgColor = termColors[c];
      curAttr |= ATTR_BGCOLOR;
    }
  }

  public void sendBytes(byte[] b) {
    if(listener != null)
      listener.sendBytes(b);
  }

  public void signalWindowChanged(int rows, int cols, int vpixels, int hpixels) {
    if(DEBUG) System.out.println("SIGWINCH: " + rows + ", " + cols);
    if(listener != null)
      listener.signalWindowChanged(rows, cols, vpixels, hpixels);
  }

  public String terminalType() {
    return interpreter.terminalType();
  }
  public int rows() {
    return rows;
  }
  public int cols() {
    return cols;
  }
  public int vpixels() {
    return vpixels;
  }
  public int hpixels() {
    return hpixels;
  }

  protected final void makeAllDirty() {
    dirtyCount = UPDATE_DIRTY_AREA;
    updateDirtyArea(0, 0, rows, cols);
    repaintPending = true;
    repaint();
  }

  protected final void updateDirtyArea(int top, int left, int bottom, int right) {
    if(top < dirtyTop)
      dirtyTop = top;
    if(bottom > dirtyBottom)
      dirtyBottom = bottom;
    if(left < dirtyLeft)
      dirtyLeft = left;
    if(right > dirtyRight)
      dirtyRight = right;
  }

  synchronized public void write(char c) {

    if(visTop != saveLines && autoRepositionOutput) {
      dirtyCount     = UPDATE_DIRTY_AREA;
      repaintPending = false; // To make the code below do repaint()
      visTop         = saveLines;
      if(scrollbar != null)
	scrollbar.setValue(visTop);
      updateDirtyArea(0, 0, rows, cols);
    }

    int ic;
    if((ic = interpreter.interpretChar(c)) != TerminalInterpreter.IGNORE) {
      c = (char)ic;
      if(curCol == cols) {
	if(lineWrap) {
	  curRow += 1;
	  curCol = 0;
	  if(curRow == windowBottom) {
	    scrollUp(1);
	    curRow = windowBottom - 1;
	  }
	} else
	  curCol--;
      }

      // Keep track of the spanning update area
      //
      dirtyCount++;
      lastWriteRow = curRow;
      lastWriteCol = curCol;
      updateDirtyArea(curRow, curCol, curRow + 1, curCol + 1);

      attributes[visTop + curRow][(curCol * IDX_LAST) + IDX_ATTR] = curAttr;
      screen[visTop + curRow][curCol++]                           = c;

    }
    if(curRow == windowBottom) {
      scrollUp(1);
      curRow = windowBottom - 1;
    }

    if(!repaintPending) {
      repaintPending = true;
      repaint();
    }
  }

  public void write(char[] c, int off, int len) {
    int end = off + len;
    for(int i = off; i < end; i++)
      write(c[i]);
  }

  public void write(String s) {
    char[] carr = s.toCharArray();
    write(carr, 0, carr.length);
  }

  public void addTerminalListener(TerminalListener listener) {
    this.listener = listener;
  }

  public void typedChar(char c) {
    // !!! used for debugging...
  }

  public void keyKludgeFilter(KeyEvent e) {
    char c = e.getKeyChar();

    if(c > (char)255) {
      switch((int)c) {
      case 65107:
	// !!! unicode-tilde (~) (on linux anyway), I know this is not
	// right, especially given tilde's use as an accent, but hey I
	// never use it as such anyway, I just want to be able to
	// print it :-) !!!
	e.setKeyChar((char)126);
	break;
      default:
	e.setKeyChar((char)0);
      }
    }

    if(enterKeyKludge) {
      if(c == 0x0a)
	e.setKeyChar((char)0x0d);
    } else if(bsKeyKludge) {
      // This should really be done with 'stty erase'/TTY_OP in SSH but I can't get
      // it working right anyway, so until then...
      //
      if(c == 0x08)
	e.setKeyChar((char)127);
    } else if(ctrlKeyKludge) {
      if(c == 0x20)
	e.setKeyChar((char)0);
    }

    if(metaKeyKludge && listener != null)
      listener.typedChar((char)27);

  }

  public void keyTyped(KeyEvent e) {
    //
    // The KeyEvent content seems to be a bit confusing in some
    // situations given different locale's and especially different
    // platforms... This is not very funny, but then again who said
    // anything about terminal-stuff beeing some kind of amusement...
    //
    keyKludgeFilter(e);

    char c = e.getKeyChar();

    if(localEcho)
      write(c);
    if(listener != null)
      listener.typedChar(c);
    else
      typedChar(c);

    if(visTop != saveLines && autoRepositionInput) {
      visTop         = saveLines;
      if(scrollbar != null)
	scrollbar.setValue(visTop);
      makeAllDirty();
    }
  }

  public void keyPressed(KeyEvent e) {
    switch(e.getKeyCode()) {
    case KeyEvent.VK_ENTER:
      enterKeyKludge = true;
      break;
    case KeyEvent.VK_ALT:
      if(ctrlKeyKludge)
	ctrlKeyKludge = false; // !!! Seems that MS-lost sends ctrl+alt for right ALT !!!
      else
	metaKeyKludge = true;
      break;
    case KeyEvent.VK_CONTROL:
      ctrlKeyKludge = true;
      break;
    case KeyEvent.VK_BACK_SPACE:
      bsKeyKludge = true;
      break;
    }
  }

  public void keyReleased(KeyEvent e) {
    String specialKey = null;
    switch(e.getKeyCode()) {
    case KeyEvent.VK_ENTER:
      enterKeyKludge = false;
      break;
    case KeyEvent.VK_ALT:
      metaKeyKludge = false;
      break;
    case KeyEvent.VK_CONTROL:
      ctrlKeyKludge = false;
      break;
    case KeyEvent.VK_BACK_SPACE:
      bsKeyKludge = false;
      break;
    case KeyEvent.VK_UP:
      if(cursorKeysMode)
	specialKey = "\033OA";
      else
	specialKey = "\033[A";
      break;
    case KeyEvent.VK_DOWN:
      if(cursorKeysMode)
	specialKey = "\033OB";
      else
	specialKey = "\033[B";
      break;
    case KeyEvent.VK_RIGHT:
      if(cursorKeysMode)
	specialKey = "\033OC";
      else
	specialKey = "\033[C";
      break;
    case KeyEvent.VK_LEFT:
      if(cursorKeysMode)
	specialKey = "\033OD";
      else
	specialKey = "\033[D";
      break;
    case KeyEvent.VK_PAGE_UP:
      pageUp();
      break;
    case KeyEvent.VK_PAGE_DOWN:
      pageDown();
      break;
    case KeyEvent.VK_HOME:
      pageHome();
      break;
    case KeyEvent.VK_END:
      pageEnd();
      break;
    }

    if(specialKey != null) {
      if(localEcho)
	write(specialKey);
      sendBytes(specialKey.getBytes());
    }
  }

  public void componentResized(ComponentEvent e) {
    Dimension dim = getSize();
    int newCols = (dim.width  - (2 * borderWidth))  / charWidth;
    int newRows = (dim.height - (2 * borderHeight)) / charHeight;
    int oldCols = cols;
    int oldRows = rows;
    char[][] oldScreen     = screen;
    int[][]  oldAttributes = attributes;

    vpixels = dim.height;
    hpixels = dim.width;

    if(newCols != oldCols)
      clearSelection();

    setWindowSize(newRows, newCols);
    resetWindow(); // !!! Is this right?
    clearScreen();

    oldCols = (oldCols < newCols ? oldCols : newCols);
    if(resizeGravity == GRAVITY_NORTHWEST) {
      int copyRows = (oldRows < newRows ? oldRows : newRows) + saveLines;
      for(int i = 0; i < copyRows; i++) {
	System.arraycopy(oldScreen[i], 0, screen[i], 0, oldCols);
	System.arraycopy(oldAttributes[i], 0, attributes[i], 0, oldCols * IDX_LAST);
      }
    } else {
      if(hasSelection) {
	selectRowAnchor += newRows - oldRows;
	selectRowLast   += newRows - oldRows;
      }
      int i, copyRows, fromTop, toTop;
      if(oldRows < newRows) {
	copyRows = oldRows + saveLines;
	fromTop  = 0;
	toTop    = newRows - oldRows;
	curRow  += toTop;
      } else {
	copyRows = newRows + saveLines;
	fromTop  = oldRows - newRows;
	toTop    = 0;
	curRow  -= fromTop;
      }
      for(i = 0; i < copyRows; i++) {
	System.arraycopy(oldScreen[i + fromTop], 0, screen[i + toTop], 0, oldCols);
	System.arraycopy(oldAttributes[i + fromTop], 0, attributes[i + toTop], 0, oldCols * IDX_LAST);
      }
    }
    if(curRow >= newRows)
      curRow = newRows - 1;
    if(curCol >= newCols)
      curCol = newCols - 1;

    updateScrollbarValues();

    signalWindowChanged(rows, cols, vpixels, hpixels);

    memGraphics = null;

    repaintPending = false; // !!! Pretend we are beeing redrawn from scratch
    repaint();
  }

  public void pageUp() {
    visTop -= rows;
    if(visTop < 0)
      visTop = 0;
    updateScrollbarValues();
    makeAllDirty();
  }
  public void pageDown() {
    visTop += rows;
    if(visTop > saveLines)
      visTop = saveLines;
    updateScrollbarValues();
    makeAllDirty();
  }
  public void pageHome() {
    visTop = 0;
    updateScrollbarValues();
    makeAllDirty();
  }
  public void pageEnd() {
    visTop = saveLines;
    updateScrollbarValues();
    makeAllDirty();
  }
  public void adjustmentValueChanged(AdjustmentEvent e) {
    int adjValue = e.getValue();
    if(adjValue >= 0 && adjValue <= saveLines) {
      visTop = adjValue;
      makeAllDirty();
    }
  }

  public void makeSelection(int startRow, int startCol, int endRow, int endCol) {
    int i, j;
    if(startRow != endRow) {
      for(i = startCol; i < cols; i++)
	attributes[startRow][(i * IDX_LAST) + IDX_ATTR] |= ATTR_SELECTED;
      for(i = startRow + 1; i < endRow; i++)
	for(j = 0; j < cols; j++)
	  attributes[i][(j * IDX_LAST) + IDX_ATTR] |= ATTR_SELECTED;
      for(i = 0; i <= endCol; i++)
	attributes[endRow][(i * IDX_LAST) + IDX_ATTR] |= ATTR_SELECTED;
    } else {
      for(i = startCol; i <= endCol; i++)
	attributes[startRow][(i * IDX_LAST) + IDX_ATTR] |= ATTR_SELECTED;
    }

    startRow -= visTop;
    endRow   -= visTop;
    if(startRow < 0)
      startRow = 0;
    if(endRow < 0)
      endRow = 0;

    updateDirtyArea(startRow, 0, endRow + 1, cols);
    if(!repaintPending) {
      dirtyCount     = UPDATE_DIRTY_AREA;
      repaintPending = true;
      repaint();
    }
  }
  public void clearSelection(int startRow, int startCol, int endRow, int endCol) {
    int i, j;
    if(startRow != endRow) {
      for(i = startCol; i < cols; i++)
	attributes[startRow][(i * IDX_LAST) + IDX_ATTR] &= ~ATTR_SELECTED;
      for(i = startRow + 1; i < endRow; i++)
	for(j = 0; j < cols; j++)
	  attributes[i][(j * IDX_LAST) + IDX_ATTR] &= ~ATTR_SELECTED;
      for(i = 0; i <= endCol; i++)
	attributes[endRow][(i * IDX_LAST) + IDX_ATTR] &= ~ATTR_SELECTED;
    } else {
      for(i = startCol; i <= endCol; i++)
	attributes[startRow][(i * IDX_LAST) + IDX_ATTR] &= ~ATTR_SELECTED;
    }

    startRow -= visTop;
    endRow   -= visTop;
    if(startRow < 0)
      startRow = 0;
    if(endRow < 0)
      endRow = 0;

    updateDirtyArea(startRow, 0, endRow + 1, cols);
    if(!repaintPending) {
      dirtyCount     = UPDATE_DIRTY_AREA;
      repaintPending = true;
      repaint();
    }
  }

  public void clearSelection() {
    if(!hasSelection)
      return;
    if(selectReverse)
      clearSelection(selectRowLast, selectColLast, selectRowAnchor, selectColAnchor);
    else
      clearSelection(selectRowAnchor, selectColAnchor, selectRowLast, selectColLast);
    hasSelection = false;
  }
  public void mousePressed(MouseEvent e) {
    int mouseRow = (e.getY() - borderHeight) / charHeight;
    int mouseCol = (e.getX() - borderWidth)  / charWidth;
    if(mouseRow < 0)
      mouseRow = 0;
    else if(mouseRow >= rows)
      mouseRow = rows - 1;
    if(mouseCol < 0)
      mouseCol = 0;
    else if(mouseCol >= cols)
      mouseCol = cols - 1;

    mouseRow += visTop;

    clearSelection();
    selectRowAnchor = mouseRow;
    selectColAnchor = mouseCol;
    selectRowLast   = mouseRow;
    selectColLast   = mouseCol;
  }
  synchronized public void mouseDragged(MouseEvent e) {
    int mouseRow = (e.getY() - borderHeight) / charHeight;
    int mouseCol = (e.getX() - borderWidth)  / charWidth;

    if(mouseRow < 0)
      mouseRow = 0;
    else if(mouseRow >= rows)
      mouseRow = rows - 1;
    if(mouseCol < 0)
      mouseCol = 0;
    else if(mouseCol >= cols)
      mouseCol = cols - 1;

    mouseRow += visTop;

    if(mouseRow == selectRowLast && mouseCol == selectColLast)
      return;

    boolean backwardSelection;
    boolean backwardsFromLast;

    if(selectRowAnchor > mouseRow ||
       (selectRowAnchor == mouseRow && mouseCol < selectColAnchor))
      backwardSelection = true;
    else
      backwardSelection = false;

    if(backwardSelection != selectReverse) {
      if(selectReverse)
	clearSelection(selectRowLast, selectColLast, selectRowAnchor, selectColAnchor);
      else
	clearSelection(selectRowAnchor, selectColAnchor, selectRowLast, selectColLast);
      selectReverse = backwardSelection;
      selectRowLast = selectRowAnchor;
      selectColLast = selectColAnchor;
    }

    if(selectRowLast > mouseRow ||
       (selectRowLast == mouseRow && mouseCol < selectColLast))
      backwardsFromLast = true;
    else
      backwardsFromLast = false;

    if(selectReverse) {
      if(backwardsFromLast)
	makeSelection(mouseRow, mouseCol, selectRowLast, selectColLast);
      else
	clearSelection(selectRowLast, selectColLast, mouseRow, mouseCol);
    } else {
      if(backwardsFromLast)
	clearSelection(mouseRow, mouseCol, selectRowLast, selectColLast);
      else
	makeSelection(selectRowLast, selectColLast, mouseRow, mouseCol);
    }

    selectReverse = backwardSelection;

    selectRowLast = mouseRow;
    selectColLast = mouseCol;

    if(selectRowAnchor == selectRowLast && selectColAnchor == selectColLast)
      hasSelection = false;
    else
      hasSelection = true;
  }

  public String getSelection() {
    if(!hasSelection)
      return null;

    int startRow;
    int endRow;
    int startCol;
    int endCol;
    int i, j;

    if(selectReverse) {
      startRow = selectRowLast;
      startCol = selectColLast;
      endRow   = selectRowAnchor;
      endCol   = selectColAnchor;
    } else {
      startRow = selectRowAnchor;
      startCol = selectColAnchor;
      endRow   = selectRowLast;
      endCol   = selectColLast;
    }

    String result = "";

    if(startRow != endRow) {
      for(i = startCol; i < cols; i++) {
	if(screen[startRow][i] == 0) {
	  i = cols;
	  break;
	}
	result += screen[startRow][i];
      }
      if(i == cols)
	result += '\n';
      for(i = startRow + 1; i < endRow; i++) {
	for(j = 0; j < cols; j++) {
	  if(screen[i][j] == 0)
	    break;
	  result += screen[i][j];
	}
	result += '\n';
      }
      for(i = 0; i <= endCol; i++) {
	if(screen[endRow][i] == 0) {
	  i = cols;
	  break;
	}
	result += screen[endRow][i];
      }
      if(i == cols)
	result += '\n';
    } else {
      for(i = startCol; i <= endCol; i++) {
	if(screen[startRow][i] == 0) {
	  i = cols;
	  break;
	}
	result += screen[startRow][i];
      }
      if(i == cols)
	result += '\n';
    }

    return result;
  }

  // !!! TODO: Move the char* init stuff out of this to a separate method
  //
  public Dimension getDimensionOfText(int rows, int cols) {
    FontMetrics fm = getFontMetrics(getFont());
    charWidth      = - 1; // !!! Does not seem to work: fm.getMaxAdvance();
    charHeight     = fm.getHeight();
    charMaxAscent  = fm.getMaxAscent();
    charMaxDescent = fm.getMaxDescent();
    charLeading    = fm.getLeading();

    if(charWidth == -1)
      charWidth = fm.charWidth('W');

    return new Dimension((cols * charWidth) + (2 * borderHeight),
			 (rows * charHeight) + (2 * borderWidth));
  }

  public Dimension getPreferredSize() {
    return getDimensionOfText(rows, cols);
  }

  public Dimension getMinimumSize() {
    return getDimensionOfText(MIN_ROWS, MIN_COLS);
  }

  protected void clearDirtyClip(Graphics g) {
    int x = borderWidth  + (charWidth   * dirtyLeft);
    int y = borderHeight + (dirtyTop    * charHeight);
    int w = (charWidth   * (dirtyRight  - dirtyLeft));
    int h = (charHeight  * (dirtyBottom - dirtyTop));
    g.setColor(origBgColor);
    g.fillRect(x, y, w, h);
    g.setColor(origFgColor);
    g.clipRect(x, y, w, h);
  }

  public void paint(Graphics g) {
    update(g);
  }

  synchronized public void update(Graphics g) {
    Dimension curSize = getSize();
    int baselineIndex = charMaxAscent + charLeading - 1;
    int x;
    int y;
    int attr;

    if(!repaintPending) {
      // If we we don't have a repaint pending the cause for update must be "destroyed"
      // window content.
      //
      dirtyCount  = UPDATE_DIRTY_AREA;
      dirtyTop    = 0;
      dirtyBottom = rows;
      dirtyLeft   = 0;
      dirtyRight  = cols;
      g.setClip(0, 0, hpixels, vpixels);
    } else {
      // Indicate that we have repainted
      //
      repaintPending = false;
    }

    // Hide last drawn cursor
    //
    x = borderWidth  + (charWidth * lastCursorCol);
    y = borderHeight + (lastCursorRow * charHeight);
    g.setColor(origBgColor);
    g.setXORMode(origFgColor);
    g.fillRect(x, y, charWidth, charHeight);
    g.setColor(origFgColor);
    g.setPaintMode();

    if(dirtyCount > 1) {
      dirtyCount  = 0;
      if(dirtyTop == rows)
	dirtyTop = curRow;
      if(dirtyBottom == 0)
	dirtyBottom = curRow + 1;
      if(dirtyRight == 0)
	dirtyRight = curCol + 1;
      if(dirtyLeft == cols)
	dirtyLeft = curCol;

      // Default is to draw directly to screen
      //
      Graphics primRenderTarget = g;

      // If the area which is to be updated is larger than MAJOR_DIRTY_AREA we do
      // "double-buffering"
      //
      boolean  majorDirty = (((dirtyRight - dirtyLeft) * (dirtyBottom - dirtyTop)) >
			     MAJOR_DIRTY_AREA);
      if(majorDirty) {
	if((memGraphics == null) ||
	   (curSize.width != memImageSize.width) ||
	   (curSize.height != memImageSize.height)) {
	  memImageSize = curSize;
	  memImage     = createImage(curSize.width, curSize.height);
	  memGraphics  = memImage.getGraphics();
	}
	clearDirtyClip(memGraphics);
	primRenderTarget = memGraphics;
      } else {
	clearDirtyClip(g);
      }
      
      for(int i = dirtyTop; i < dirtyBottom; i++) {
	y = borderHeight + (i * charHeight);
	for(int j = dirtyLeft; j < dirtyRight; j++) {
	  attr = attributes[visTop + i][(j * IDX_LAST) + IDX_ATTR];
	  // Skip positions not drawn into
	  if(attr == ATTR_CHARNOTDRAWN)
	    continue;
	  x = borderWidth  + (charWidth * j);
	  if(((attr & ATTR_INVERSE) != 0) ^ ((attr & ATTR_SELECTED) != 0)) {
	    primRenderTarget.fillRect(x, y, charWidth, charHeight);
	    primRenderTarget.setColor(origBgColor);
	  }
	  if(attr != ATTR_SELECTED) {
	    if((attr & ATTR_BOLD) != 0) {
	      primRenderTarget.setFont(boldFont);
	      primRenderTarget.drawChars(screen[visTop + i], j, 1, x, y + baselineIndex);
	      primRenderTarget.setFont(plainFont);
	    } else {
	      primRenderTarget.drawChars(screen[visTop + i], j, 1, x, y + baselineIndex);
	    }
	    if((attr & ATTR_UNDERLINE) != 0)
	      primRenderTarget.drawLine(x, y + baselineIndex, x + charWidth, y + baselineIndex);
	  }
	  primRenderTarget.setColor(origFgColor);
	}
      }

      if(majorDirty) {
	clearDirtyClip(g);
	g.drawImage(memImage, 0, 0, this);
	memGraphics.setClip(0, 0, hpixels, vpixels);
      }
    } else if(dirtyCount == 1) {
      dirtyCount = 0;
      attr = attributes[visTop + lastWriteRow][(lastWriteCol * IDX_LAST) + IDX_ATTR];
      x    = borderWidth  + (charWidth * lastWriteCol);
      y    = borderHeight + (lastWriteRow * charHeight);
      if((attr & ATTR_INVERSE) != 0) {
	g.fillRect(x, y, charWidth, charHeight);
	g.setColor(origBgColor);
      } else {
	g.setColor(origBgColor);
	g.fillRect(x, y, charWidth, charHeight);
	g.setColor(origFgColor);
      }
      if((attr & ATTR_BOLD) != 0) {
	g.setFont(boldFont);
	g.drawChars(screen[visTop + lastWriteRow], lastWriteCol, 1, x, y + baselineIndex);
	g.setFont(plainFont);
      } else {
	g.drawChars(screen[visTop + lastWriteRow], lastWriteCol, 1, x, y + baselineIndex);
      }
      if((attr & ATTR_UNDERLINE) != 0)
	g.drawLine(x, y + baselineIndex, x + charWidth, y + baselineIndex);
      g.setColor(origFgColor);
    }

    // Reset dirty area (i.e. nothing is dirty now)
    //
    dirtyTop    = rows;
    dirtyBottom = 0;
    dirtyLeft   = cols;
    dirtyRight  = 0;
    g.setClip(0, 0, hpixels, vpixels);

    // Show current cursor
    //
    x = borderWidth  + (charWidth * curCol);
    y = borderHeight + (curRow * charHeight);
    g.setColor(origFgColor);
    g.setXORMode(origBgColor);
    g.fillRect(x, y, charWidth, charHeight);
    g.setPaintMode();

    lastCursorRow = curRow;
    lastCursorCol = curCol;
  }

}
